#!/usr/bin/python2.7
# duplicity-hardlinks.py -- Extend duplicity functionality to hard links
#
# Made by Mihail Shapochkin <shivaa@mail.ru> using rdiffdir as a template.
#
# This file should be a part of duplicity! :)
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# See http://www.nongnu.org/duplicity for more information.
# Please send mail to me or the mailing list if you find bugs or have
# any suggestions.


import sys, getopt, os

import gettext
gettext.install('duplicity', codeset='utf8')

from duplicity import log
from duplicity import globals
from duplicity import selection
from duplicity import path

select_opts = [] # Add selection argument tuples to this
select_files = [] # Will hold file objects when filelist given

def parse_cmdline_options(arglist):
    """Parse argument list"""
    global select_opts, select_files
    def sel_fl(filename):
        """Helper function for including/excluding filelists below"""
        try:
            return open(filename, "r")
        except IOError:
            log.FatalError("Error opening file %s" % filename)

    try:
        optlist, args = getopt.getopt(arglist, "v:Vz",
                                      ["exclude=", "exclude-device-files",
                                       "exclude-filelist=", "exclude-filelist-stdin",
                                       "exclude-globbing-filelist=", "exclude-other-filesystems",
                                       "exclude-regexp=", "include=", "include-filelist=",
                                       "include-filelist-stdin", "include-globbing-filelist=",
                                       "include-regexp=", "null-separator", "verbosity="])
    except getopt.error, e:
        command_line_error("Bad command line option: %s" % (str(e),))

    for opt, arg in optlist:
        if (opt == "--exclude" or opt == "--exclude-regexp" or
              opt == "--include" or opt == "--include-regexp"):
            select_opts.append((opt, arg))
        elif (opt == "--exclude-device-files" or
              opt == "--exclude-other-filesystems"):
            select_opts.append((opt, None))
        elif (opt == "--exclude-filelist" or opt == "--include-filelist" or
              opt == "--exclude-globbing-filelist" or
              opt == "--include-globbing-filelist"):
            select_opts.append((opt, arg))
            select_files.append(sel_fl(arg))
        elif opt == "--exclude-filelist-stdin":
            select_opts.append(("--exclude-filelist", "standard input"))
            select_files.append(sys.stdin)
        elif opt == "--include-filelist-stdin":
            select_opts.append(("--include-filelist", "standard input"))
            select_files.append(sys.stdin)
        elif opt == "--null-separator":
            globals.null_separator = 1
        elif opt == "-V":
            print "duplicity-hardlinks.py", str(globals.version)
            sys.exit(0)
        elif opt == "-v" or opt == "--verbosity":
            log.setverbosity(int(arg))
        else:
            command_line_error("Unknown option %s" % opt)

    return args

def command_line_error(message):
    """Indicate a command line error and exit"""
    sys.stderr.write("Error: %s\n" % (message,))
    sys.stderr.write("Usage:\n\tduplicity-hardlinks.py [options] dump source_dir dump_file\n\tduplicity-hardlinks.py restore dump_file target_dir\n")
    sys.stderr.write("\t[options] - the syntax is similar to duplicity/rdiff\n")
    sys.stderr.write("Example:\n")
    sys.stderr.write("\tduplicity-hardlinks.py --include /home --include /etc --exclude '**' dump / /hardlinks_dump.txt\n")
    sys.stderr.write("\tThis will save dump of all hardlinks from /home and /etc directories to file /hardlinks_dump.txt.\n")
    sys.stderr.write("\tThen use duplicity to backup your files including your /hardlinks_dump.txt file!\n")
    sys.stderr.write("\tRestore ordinal duplicity backup to some directory, for example /mnt/restore/\n\tFile hardlinks_dump.txt should be there.\n")
    sys.stderr.write("\tNow restore the hardlinks structure:\n")
    sys.stderr.write("\tduplicity-hardlinks.py restore hardlinks_dump.txt /mnt/restore\n")
    sys.exit(1)

def check_does_not_exist(filename):
    """Exit with error message if filename already exists"""
    try:
        os.lstat(filename)
    except OSError:
        pass
    else:
        #log.FatalError("File %s already exists, will not "
        #               "overwrite." % filename)
        log.Warn("File %s already exists, will "
                       "overwrite." % filename)

def get_action(args):
    """Figure out the main action from the arguments"""
    def require_args(num):
        if len(args)-1 < num:
            command_line_error("Too few arguments")
        elif len(args)-1 > num:
            command_line_error("Too many arguments")

    if not args:
        command_line_error("No arguments found")
    command = args[0]
    if command == "dump":
        require_args(2)
    elif command == "restore":
        require_args(2)
    return command, args[1:]

def get_selection(filename):
    """Return selection iter starting at path with arguments applied"""
    global select_opts, select_files
    sel = selection.Select(path.Path(filename))
    sel.ParseArgs(select_opts, select_files)
    return sel.set_iter()

def get_fileobj(filename, mode):
    """Get file object or stdin/stdout from filename"""
    if mode == "r" or mode == "rb":
        if filename == "-":
            fp = sys.stdin
        else:
            fp = open(filename, mode)
    elif mode == "w" or mode == "wb":
        if filename == "-":
            fp = sys.stdout
        else:
            check_does_not_exist(filename)
            fp = open(filename, mode)
    else:
        assert 0, "Unknown mode " + str(mode)

    return fp

def dump_hardlinks(dirname, outfp):
    """Write dump of hardlinks from dirname into file object outfp"""
    hardlinks = dict();
    for file in get_selection(dirname):
        # during restore should do: file1.compare_data(file2), file1.perms_equal(file2)
        if file.type == "reg" and file.stat.st_nlink > 1:
            log.Debug( "Found hard link: %s" % file.name )
            #print file.name
            #print file.base
            #print file.index
            #print file.stat
            #print file.get_relative_path()
            key = str(file.stat.st_dev) + " " + str(file.stat.st_ino); # should include device id
            if hardlinks.has_key(key):
                hardlinks[key].append(file.get_relative_path())
            else:
                hardlinks[key] = [ file.get_relative_path() ];

    # done, saving
    counter = 0
    for k,hl in hardlinks.items():
        if len(hl) > 1:
            outfp.write("\n".join(hl) + "\n\n") # simple format
            counter += len(hl)-1 # -1 to make numbers of dump/restore equal

    print "%u hard links dumped" % counter

    assert not outfp.close()
        
        
def restore_hardlinks(infp, dirname):
    """Read dump of hardlinks from infp and apply to files in dirname"""
    hardlinks = [];
    hl = []
    for line in iter(infp):
        if line == "\n":
            if len(hl) > 1:
                hardlinks.append(hl)
            hl = []
        else:
            file_path = path.Path(dirname+"/"+line.rstrip())
            if file_path.stat == None:
                log.Notice( "File %s is absent, skipping, partial restore ?" % file_path.base )
            else:
                hl.append(file_path)
    if len(hl) > 1:
        hardlinks.append(hl)
    infp.close()


    counter = 0
    for hl in hardlinks:
        first_file = hl.pop()
        for file_path in hl:
            # check if files are equal, croak if not.
            #file1.compare_data(file2), file1.perms_equal(file2)
            if not first_file.compare_data(file_path):
                log.Warn( "Files %s and %s are not equal, skipping." % (first_file.base, file_path.base) )
            else:
                command =  "ln --force \"%s\" \"%s\"" % (first_file.base, file_path.base)
                log.Debug( "Running command: %s" % command );
                #os.link(first_file.base, file_path.base); - do not use it! so many points of failure here, system ln is much more reliable
                os.system(command)
                counter += 1
    
    print "%u hard links restored" % counter
            


def main():
    """Start here"""
    log.setup()
    args = parse_cmdline_options(sys.argv[1:])
    action, file_args = get_action(args)
    if action == "dump":
        os.umask(0077) # list of files can be secret, force 0600 permissions
        dump_hardlinks(file_args[0], get_fileobj(file_args[1], "wb"))
    elif action == "restore":
        restore_hardlinks(get_fileobj(file_args[0], "rb"), file_args[1])
    else:
        command_line_error("Bad command " + action)


if __name__ == "__main__":
    main()
